# coding: utf-8


# docs
from __future__ import annotations

from typing import Dict, List, Optional, Tuple, Union

import cv2
import numpy as np
from shapely.geometry import Polygon as ShapelyPolygon
from shapely.geometry import mapping

from supervisely.geometry import validation
from supervisely.geometry.constants import (
    CLASS_ID,
    CREATED_AT,
    EXTERIOR,
    ID,
    INTERIOR,
    LABELER_LOGIN,
    POINTS,
    UPDATED_AT,
)
from supervisely.geometry.conversions import shapely_figure_to_coords_list
from supervisely.geometry.point_location import (
    PointLocation,
    points_to_row_col_list,
    row_col_list_to_points,
)
from supervisely.geometry.rectangle import Rectangle
from supervisely.geometry.vector_geometry import VectorGeometry
from supervisely.sly_logger import logger


class Polygon(VectorGeometry):
    """
    Polygon geometry for a single :class:`Label<supervisely.annotation.label.Label>`. :class:`Polygon<Polygon>` class object is immutable.

    :param exterior: Exterior coordinates, object contour is defined with these points.
    :type exterior: List[PointLocation], List[List[int, int]], List[Tuple[int, int]
    :param interior: Interior coordinates, object holes are defined with these points.
    :type interior: List[List[PointLocation]], List[List[List[int, int]]], List[List[Tuple[int, int]]]
    :param sly_id: Polygon ID in Supervisely server.
    :type sly_id: int, optional
    :param class_id: ID of :class:`ObjClass<supervisely.annotation.obj_class.ObjClass>` to which Polygon belongs.
    :type class_id: int, optional
    :param labeler_login: Login of the user who created Polygon.
    :type labeler_login: str, optional
    :param updated_at: Date and Time when Polygon was modified last. Date Format: Year:Month:Day:Hour:Minute:Seconds. Example: '2021-01-22T19:37:50.158Z'.
    :type updated_at: str, optional
    :param created_at: Date and Time when Polygon was created. Date Format is the same as in "updated_at" parameter.
    :type created_at: str, optional
    :raises: :class:`ValueError`, if len(exterior) < 3 or len(any element in interior list) < 3

    :Usage example:

     .. code-block:: python

            import supervisely as sly

            exterior = [sly.PointLocation(730, 2104), sly.PointLocation(2479, 402), sly.PointLocation(3746, 1646)]
            # or exterior = [[730, 2104], [2479, 402], [3746, 1646]]
            # or exterior = [(730, 2104), (2479, 402), (3746, 1646)]
            interior = [[sly.PointLocation(1907, 1255), sly.PointLocation(2468, 875), sly.PointLocation(2679, 1577)]]
            # or interior = [[[730, 2104], [2479, 402], [3746, 1646]]]
            # or interior = [[(730, 2104), (2479, 402), (3746, 1646)]]
            figure = sly.Polygon(exterior, interior)
    """

    @staticmethod
    def geometry_name():
        """ """
        return "polygon"

    def __init__(
        self,
        exterior: Union[List[PointLocation], List[List[int, int]], List[Tuple[int, int]]],
        interior: Union[
            List[List[PointLocation]], List[List[List[int, int]]], List[List[Tuple[int, int]]]
        ] = [],
        sly_id: Optional[int] = None,
        class_id: Optional[int] = None,
        labeler_login: Optional[int] = None,
        updated_at: Optional[str] = None,
        created_at: Optional[str] = None,
    ):
        if len(exterior) < 3:
            exterior.extend([exterior[-1]] * (3 - len(exterior)))
            logger.warn(
                f'"{EXTERIOR}" field must contain at least 3 points to create "Polygon" object.'
            )
            # raise ValueError('"{}" field must contain at least 3 points to create "Polygon" object.'.format(EXTERIOR))
        for element in interior:
            if len(element) < 3:
                logger.warn(
                    f'"{element}" interior field must contain at least 3 points to create "Polygon" object.'
                )
                element.extend([element[-1]] * (3 - len(element)))
        # if any(len(element) < 3 for element in interior):
        #    raise ValueError('"{}" element must contain at least 3 points.'.format(INTERIOR))

        super().__init__(
            exterior,
            interior,
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    @classmethod
    def from_json(cls, data: Dict) -> Polygon:
        """
        Convert a json dict to Polygon. Read more about `Supervisely format <https://docs.supervisely.com/data-organization/00_ann_format_navi>`_.

        :param data: Polygon in json format as a dict.
        :type data: dict
        :return: Polygon object
        :rtype: :class:`Polygon<Polygon>`
        :Usage example:

         .. code-block:: python

            import supervisely as sly

            figure_json =  {
                "points": {
                    "exterior": [
                        [2104, 730],
                        [402, 2479],
                        [1646, 3746]
                    ],
                    "interior": [
                        [
                            [1255, 1907],
                            [875, 2468],
                            [577, 2679]
                        ]
                    ]
                }
            }

            figure = sly.Polygon.from_json(figure_json)
        """
        validation.validate_geometry_points_fields(data)
        labeler_login = data.get(LABELER_LOGIN, None)
        updated_at = data.get(UPDATED_AT, None)
        created_at = data.get(CREATED_AT, None)
        sly_id = data.get(ID, None)
        class_id = data.get(CLASS_ID, None)
        return cls(
            exterior=row_col_list_to_points(data[POINTS][EXTERIOR], flip_row_col_order=True),
            interior=[
                row_col_list_to_points(i, flip_row_col_order=True) for i in data[POINTS][INTERIOR]
            ],
            sly_id=sly_id,
            class_id=class_id,
            labeler_login=labeler_login,
            updated_at=updated_at,
            created_at=created_at,
        )

    def crop(self, rect: Rectangle) -> List[Polygon]:
        """
        Crops current Polygon.

        :param rect: Rectangle object for crop.
        :type rect: Rectangle
        :return: List of Polygon objects
        :rtype: :class:`List[Polygon]<Polygon>`

        :Usage Example:

         .. code-block:: python

            import supervisely as sly

            crop_figures = figure.crop(sly.Rectangle(1, 1, 300, 350))
        """
        try:
            # points = [
            #     PointLocation(row=rect.top, col=rect.left),
            #     PointLocation(row=rect.top, col=rect.right + 1),
            #     PointLocation(row=rect.bottom + 1, col=rect.right + 1),
            #     PointLocation(row=rect.bottom + 1, col=rect.left)
            # ]
            points = [
                PointLocation(row=rect.top, col=rect.left),
                PointLocation(row=rect.top, col=rect.right),
                PointLocation(row=rect.bottom, col=rect.right),
                PointLocation(row=rect.bottom, col=rect.left),
            ]
            # points = rect.corners # old implementation with 1 pixel error (right bottom)
            # #@TODO: investigate here (critical issue)

            clipping_window_shpl = ShapelyPolygon(points_to_row_col_list(points))
            self_shpl = ShapelyPolygon(self.exterior_np, holes=self.interior_np)
            intersections_shpl = self_shpl.buffer(0).intersection(clipping_window_shpl)
            mapping_shpl = mapping(intersections_shpl)
        except Exception:
            logger.warn("Polygon cropping exception, shapely.", exc_info=True)
            # raise
            # if polygon is invalid, just print warning and skip it
            # @TODO: need more investigation here
            return []

        intersections = shapely_figure_to_coords_list(mapping_shpl)

        # Check for bad cropping cases (e.g. empty points list)
        out_polygons = []
        for intersection in intersections:
            if (
                isinstance(intersection, list)
                and len(intersection) > 0
                and len(intersection[0]) >= 3
            ):
                exterior = row_col_list_to_points(intersection[0], do_round=True)
                interiors = []
                for interior_contour in intersection[1:]:
                    if len(interior_contour) > 2:
                        interiors.append(row_col_list_to_points(interior_contour, do_round=True))
                out_polygons.append(Polygon(exterior, interiors))
        return out_polygons

    def _draw_impl(self, bitmap, color, thickness=1, config=None):
        """ """
        exterior = self.exterior_np[:, ::-1]
        interior = [x[:, ::-1] for x in self.interior_np]
        bmp_to_draw = np.zeros(bitmap.shape[:2], np.uint8)
        cv2.fillPoly(bmp_to_draw, pts=[exterior], color=1)
        cv2.fillPoly(bmp_to_draw, pts=interior, color=0)
        bool_mask = bmp_to_draw.astype(bool)
        bitmap[bool_mask] = color

    def _draw_contour_impl(self, bitmap, color, thickness=1, config=None):
        """ """
        exterior = self.exterior_np[:, ::-1]
        interior = [x[:, ::-1] for x in self.interior_np]

        poly_lines = [exterior] + interior
        cv2.polylines(bitmap, pts=poly_lines, isClosed=True, color=color, thickness=thickness)

    @property
    def area(self) -> float:
        """
        Polygon area.

        :return: Area of current Polygon object.
        :rtype: :class:`float`

        :Usage Example:

         .. code-block:: python

            print(figure.area)
            # Output: 7288.0
        """
        exterior_area = self._get_area_by_gauss_formula(
            self.exterior_np[:, 0], self.exterior_np[:, 1]
        )
        interior_area = 0
        for hole in self.interior_np:
            interior_area += self._get_area_by_gauss_formula(hole[:, 0], hole[:, 1])
        return exterior_area - interior_area

    @staticmethod
    def _get_area_by_gauss_formula(rows, cols):
        """ """
        return 0.5 * np.abs(np.dot(rows, np.roll(cols, 1)) - np.dot(cols, np.roll(rows, 1)))

    def approx_dp(self, epsilon: float) -> Polygon:
        """
        Approximates a Polygon curve with the specified precision.

        :param epsilon: Specifying the approximation accuracy.
        :type epsilon: float
        :return: Polygon object
        :rtype: :class:`Polygon<Polygon>`

        :Usage Example:

         .. code-block:: python

            # Remember that Polygon class object is immutable, and we need to assign new instance of Polygon to a new variable
            approx_figure = figure.approx_dp(0.75)
        """
        exterior_np = self._approx_ring_dp(self.exterior_np, epsilon, closed=True).tolist()
        interior_np = [
            self._approx_ring_dp(x, epsilon, closed=True).tolist() for x in self.interior_np
        ]
        exterior = row_col_list_to_points(exterior_np, do_round=True)
        interior = [row_col_list_to_points(x, do_round=True) for x in interior_np]
        return Polygon(exterior, interior)

    @classmethod
    def allowed_transforms(cls):
        """ """
        from supervisely.geometry.alpha_mask import AlphaMask
        from supervisely.geometry.any_geometry import AnyGeometry
        from supervisely.geometry.bitmap import Bitmap
        from supervisely.geometry.rectangle import Rectangle

        return [AnyGeometry, Rectangle, Bitmap, AlphaMask]



STATE = "state"
DATA = "data"
CONTEXT = "context"
TEMPLATE = "template"

SHARED_DATA = '/sessions'

STOP_COMMAND = "stop"

IMAGE_ANNOTATION_EVENTS = ["manual_selected_figure_changed"]

# Error message for missing or incompatible protobuf dependencies
PROTOBUF_REQUIRED_ERROR = (
    "protobuf is required for agent/worker/app_v1 functionality. "
    "Please install supervisely with agent extras: pip install 'supervisely[agent]'"
)



# Pointcloud Episodes
from supervisely.convert.pointcloud_episodes.sly.sly_pointcloud_episodes_converter import (
    SLYPointcloudEpisodesConverter,
)
from supervisely.convert.pointcloud_episodes.bag.bag_converter import BagEpisodesConverter
from supervisely.convert.pointcloud_episodes.lyft.lyft_converter import LyftEpisodesConverter
from supervisely.convert.pointcloud_episodes.nuscenes_conv.nuscenes_converter import (
    NuscenesEpisodesConverter,
)
from supervisely.convert.pointcloud_episodes.kitti_360.kitti_360_converter import KITTI360Converter



import hashlib
from contextlib import contextmanager
from typing import Optional, Union

from fastapi import Request

import supervisely.io.env as sly_env
from supervisely.api.module_api import ApiField
from supervisely.app.fastapi.websocket import WebsocketManager
from supervisely.sly_logger import logger


def _parse_int(value):
    try:
        return int(value)
    except (TypeError, ValueError):
        return None


def _user_identity_from_cookie(request: Request) -> Optional[str]:
    cookie_header = request.headers.get("cookie")
    if not cookie_header:
        return None
    return hashlib.sha256(cookie_header.encode("utf-8")).hexdigest()


async def extract_user_id_from_request(request: Request) -> Optional[Union[int, str]]:
    """Extract user ID from various parts of the request."""
    if not sly_env.is_multiuser_mode_enabled():
        return None
    user_id = _parse_int(request.query_params.get("userId"))
    if user_id is None:
        header_user = _parse_int(request.headers.get("x-user-id"))
        if header_user is not None:
            user_id = header_user
    if user_id is None:
        referer = request.headers.get("referer", "")
        if referer:
            from urllib.parse import parse_qs, urlparse

            try:
                parsed_url = urlparse(referer)
                query_params = parse_qs(parsed_url.query)
                referer_user = query_params.get("userId", [None])[0]
                user_id = _parse_int(referer_user)
            except Exception as e:
                logger.error(f"Error parsing userId from referer: {e}")
    if user_id is None and "application/json" in request.headers.get("Content-Type", ""):
        try:
            payload = await request.json()
        except Exception:
            payload = {}
        context = payload.get("context") or {}
        user_id = _parse_int(context.get("userId") or context.get(ApiField.USER_ID))
        if user_id is None:
            state_payload = payload.get("state") or {}
            user_id = _parse_int(state_payload.get("userId") or state_payload.get(ApiField.USER_ID))
    if user_id is None:
        user_id = _user_identity_from_cookie(request)
    return user_id


@contextmanager
def session_context(user_id: Optional[Union[int, str]]):
    """
    Context manager to set and reset user context for multiuser applications.
    Call this at the beginning of a request handling to ensure the correct user context is set in environment variables (`supervisely_multiuser_app_user_id` ContextVar).
    """
    if not sly_env.is_multiuser_mode_enabled() or user_id is None:
        yield
        return
    token = sly_env.set_user_for_multiuser_app(user_id)
    try:
        yield
    finally:
        sly_env.reset_user_for_multiuser_app(token)


def remember_cookie(request: Request, user_id: Optional[Union[int, str]]):
    """
    Remember user cookie for the given user ID. This is used to associate WebSocket connections with users in multiuser applications based on cookies.
    Allows WebSocket connections to be correctly routed to the appropriate user.
    """
    if not sly_env.is_multiuser_mode_enabled() or user_id is None:
        return
    cookie_header = request.headers.get("cookie")
    if cookie_header:
        WebsocketManager().remember_user_cookie(cookie_header, user_id)



from __future__ import annotations

import numpy as np
import pandas as pd

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class F1ScoreAtDifferentIOU(DetectionVisMetric):
    MARKDOWN = "f1_score_at_iou"
    CHART = "f1_score_at_iou"

    @property
    def md(self) -> MarkdownWidget:
        return MarkdownWidget(
            self.MARKDOWN,
            "Confidence Profile at Different IoU thresholds",
            self.vis_texts.markdown_f1_at_ious,
        )

    @property
    def chart(self) -> ChartWidget:
        return ChartWidget(self.CHART, self._get_figure())

    def _get_figure(self):  # -> go.Figure:
        import plotly.express as px  # pylint: disable=import-error

        f1s = self.eval_result.mp.m_full.score_profile_f1s

        # downsample
        if len(self.eval_result.df_score_profile) > 5000:
            f1s_down = f1s[:, :: f1s.shape[1] // 1000]
        else:
            f1s_down = f1s

        iou_names = list(map(lambda x: str(round(x, 2)), self.eval_result.mp.iouThrs.tolist()))
        df = pd.DataFrame(
            np.concatenate([self.eval_result.dfsp_down["scores"].values[:, None], f1s_down.T], 1),
            columns=["scores"] + iou_names,
        )
        labels = {"value": "F1-score", "variable": "IoU threshold", "scores": "Confidence Score"}

        fig = px.line(
            df,
            x="scores",
            y=iou_names,
            labels=labels,
            color_discrete_sequence=px.colors.sequential.Viridis,
            width=None,
            height=500,
        )
        fig.update_traces(
            hovertemplate="Confidence Score: %{x:.2f}<br>F1-score: %{y:.2f}<extra></extra>"
        )
        fig.update_layout(yaxis=dict(range=[0, 1]), xaxis=dict(range=[0, 1], tick0=0, dtick=0.1))

        # add annotations for maximum F1-Score for each IoU threshold
        for i, iou in enumerate(iou_names):
            # Skip if all F1 scores are NaN for this IoU threshold
            if np.isnan(f1s[i]).all():
                continue
            argmax_f1 = np.nanargmax(f1s[i])
            max_f1 = f1s[i][argmax_f1]
            score = self.eval_result.mp.m_full.score_profile["scores"][argmax_f1]
            fig.add_annotation(
                x=score,
                y=max_f1,
                text=f"Best conf: {score:.2f}",
                showarrow=True,
                arrowhead=1,
                arrowcolor="black",
                ax=0,
                ay=-30,
            )

        fig.update_layout(
            dragmode=False,
            modebar=dict(
                remove=[
                    "zoom2d",
                    "pan2d",
                    "select2d",
                    "lasso2d",
                    "zoomIn2d",
                    "zoomOut2d",
                    "autoScale2d",
                    "resetScale2d",
                ]
            ),
        )
        return fig



from typing import Any, Dict, List

import numpy as np

from supervisely.annotation.label import Label
from supervisely.annotation.obj_class import ObjClass
from supervisely.geometry.bitmap import Bitmap
from supervisely.nn.inference.inference import Inference
from supervisely.nn.prediction_dto import PredictionSegmentation
from supervisely.project.project_meta import ProjectMeta
from supervisely.sly_logger import logger


class SemanticSegmentation(Inference):
    def get_info(self) -> dict:
        info = super().get_info()
        info["task type"] = "semantic segmentation"
        info["tracking_on_videos_support"] = False
        # recommended parameters:
        # info["model_name"] = ""
        # info["checkpoint_name"] = ""
        # info["pretrained_on_dataset"] = ""
        # info["device"] = ""
        return info

    def _get_obj_class_shape(self):
        return Bitmap

    def _find_bg_class_index(self, class_names: List[str]):
        possible_bg_names = ["background", "bg", "unlabeled", "neutral", "__bg__"]
        bg_class_index = None
        for i, name in enumerate(class_names):
            if name in possible_bg_names:
                bg_class_index = i
                break
        return bg_class_index

    def _add_default_bg_class(self, meta: ProjectMeta):
        default_bg_class_name = "__bg__"
        obj_class = meta.get_obj_class(default_bg_class_name)
        if obj_class is None:
            obj_class = ObjClass(default_bg_class_name, self._get_obj_class_shape())
            meta = meta.add_obj_class(obj_class)
        return meta, obj_class

    def _get_or_create_bg_obj_class(self, classes):
        bg_class_index = self._find_bg_class_index(classes)
        if bg_class_index is None:
            self._model_meta, bg_obj_class = self._add_default_bg_class(self.model_meta)
        else:
            bg_class_name = classes[bg_class_index]
            bg_obj_class = self.model_meta.get_obj_class(bg_class_name)
        return bg_obj_class

    def _create_label(self, dto: PredictionSegmentation, classes_whitelist: List[str] = None):
        classes = self.get_classes()

        image_classes_indexes = np.unique(dto.mask)
        labels = []
        for class_idx in image_classes_indexes:
            class_mask = dto.mask == class_idx
            class_name = classes[class_idx]
            if classes_whitelist not in (None, "all") and class_name not in classes_whitelist:
                obj_class = self._get_or_create_bg_obj_class(classes)
            else:
                obj_class = self.model_meta.get_obj_class(class_name)
            if obj_class is None:
                raise KeyError(
                    f"Class {class_name} not found in model classes {self.get_classes()}"
                )
            geometry = Bitmap(class_mask, extra_validation=False)
            label = Label(geometry, obj_class)
            labels.append(label)
        return labels

    def predict(self, image_path: str, settings: Dict[str, Any]) -> List[PredictionSegmentation]:
        raise NotImplementedError("Have to be implemented in child class")

    def predict_raw(
        self, image_path: str, settings: Dict[str, Any]
    ) -> List[PredictionSegmentation]:
        raise NotImplementedError(
            "Have to be implemented in child class If sliding_window_mode is 'advanced'."
        )



from collections import defaultdict
from typing import Callable, List, Literal, Optional

import numpy as np

from supervisely.nn.benchmark.utils.detection.coco_eval import (
    COCO,
    SlyCOCOeval,
    pycocotools_installed,
)


def set_cocoeval_params(
    cocoeval: SlyCOCOeval,
    parameters: dict,
):
    """
    type cocoeval: COCOeval
    """
    if parameters is None:
        return
    param_names = (
        "iouThrs",
        "recThrs",
        "maxDets",
        "areaRng",
        "areaRngLbl",
        # "kpt_oks_sigmas" # For keypoints
    )
    for param_name in param_names:
        cocoeval.params.__setattr__(
            param_name, parameters.get(param_name, cocoeval.params.__getattribute__(param_name))
        )


def calculate_metrics(
    cocoGt: COCO,
    cocoDt: COCO,
    iouType: Literal["bbox", "segm"],
    progress_cb: Optional[Callable] = None,
    evaluation_params: Optional[dict] = None,
):
    """
    Calculate COCO metrics.

    :param cocoGt: Ground truth dataset in COCO format
    :type cocoGt: COCO
    :param cocoDt: Predicted dataset in COCO format
    :type cocoDt: COCO
    :param iouType: Type of IoU calculation
    :type iouType: Literal["bbox", "segm"]
    :param progress_cb: Progress callback
    :type progress_cb: Optional[Callable]
    :return: Results of the evaluation
    :rtype: dict
    """
    if not pycocotools_installed:
        raise ImportError("pycocotools is not installed")

    evaluation_params = evaluation_params or {}
    max_dets = evaluation_params.get("max_detections", 100)

    cocoEval = SlyCOCOeval(cocoGt, cocoDt, iouType=iouType)
    cocoEval.params.maxDets[-1] = max_dets
    cocoEval.evaluate()
    progress_cb(1) if progress_cb is not None else None
    cocoEval.accumulate()
    progress_cb(1) if progress_cb is not None else None
    cocoEval.summarize()

    # For classification metrics
    cocoEval_cls = SlyCOCOeval(cocoGt, cocoDt, iouType=iouType)
    cocoEval_cls.params.useCats = 0
    cocoEval_cls.params.maxDets[-1] = max_dets
    cocoEval_cls.evaluate()
    progress_cb(1) if progress_cb is not None else None
    cocoEval_cls.accumulate()
    progress_cb(1) if progress_cb is not None else None
    cocoEval_cls.summarize()

    iouThrs = cocoEval.params.iouThrs
    iou_threshold = evaluation_params.get("iou_threshold", 0.5)
    iou_threshold_per_class = evaluation_params.get("iou_threshold_per_class")
    if iou_threshold_per_class is not None:
        iou_idx_per_class = {
            cocoGt.getCatIds(catNms=[class_name])[0]: np.where(np.isclose(iouThrs, iou_thres))[0][0]
            for class_name, iou_thres in iou_threshold_per_class.items()
        }
    else:
        iou_idx = np.where(np.isclose(iouThrs, iou_threshold))[0][0]
        iou_idx_per_class = {cat_id: iou_idx for cat_id in cocoGt.getCatIds()}

    # TODO: Add support for average_across_iou_thresholds
    if iou_threshold_per_class is not None or iou_threshold != 0.5:
        average_across_iou_thresholds = False
        evaluation_params["average_across_iou_thresholds"] = average_across_iou_thresholds

    eval_img_dict = get_eval_img_dict(cocoEval)
    eval_img_dict_cls = get_eval_img_dict(cocoEval_cls)
    matches = get_matches(
        eval_img_dict,
        eval_img_dict_cls,
        cocoEval_cls,
        iou_idx_per_class=iou_idx_per_class,
    )
    # true_positives, false_positives, false_negatives = get_counts(eval_img_dict, cocoEval_cls)

    params = {
        "iouThrs": cocoEval.params.iouThrs,
        "recThrs": cocoEval.params.recThrs,
        "evaluation_params": evaluation_params,
        "iou_idx_per_class": iou_idx_per_class,
    }
    coco_metrics = {"mAP": cocoEval.stats[0], "precision": cocoEval.eval["precision"]}
    coco_metrics["AP50"] = cocoEval.stats[1]
    coco_metrics["AP75"] = cocoEval.stats[2]
    eval_data = {
        "matches": matches,
        "coco_metrics": coco_metrics,
        "params": params,
    }
    progress_cb(1) if progress_cb is not None else None

    return eval_data


def get_counts(eval_img_dict: dict, cocoEval_cls: SlyCOCOeval):
    if not pycocotools_installed:
        raise ImportError("pycocotools is not installed")

    cat_ids = cocoEval_cls.cocoGt.getCatIds()
    iouThrs = cocoEval_cls.params.iouThrs
    catId2idx = {cat_id: i for i, cat_id in enumerate(cat_ids)}
    true_positives = np.zeros((len(cat_ids), len(iouThrs)))
    false_positives = np.zeros((len(cat_ids), len(iouThrs)))
    false_negatives = np.zeros((len(cat_ids), len(iouThrs)))
    for img_id, eval_imgs in eval_img_dict.items():
        for eval_img in eval_imgs:
            cat_id = eval_img["category_id"]
            cat_idx = catId2idx[cat_id]
            gtIgnore = eval_img["gtIgnore"]
            # if conf_thresh is not None:
            #     scores = np.array(eval_img["dtScores"])
            #     dt_conf_mask = scores < conf_thresh
            #     dt_not_ignore[:, dt_conf_mask] = False
            #     for idx, dt_id in enumerate(eval_img['dtIds']):
            #         if dt_conf_mask[idx]:
            #             eval_img["gtMatches"][eval_img["gtMatches"] == dt_id] = 0.
            gt_not_ignore_idxs = np.where(np.logical_not(gtIgnore))[0]
            dt_not_ignore = np.logical_not(eval_img["dtIgnore"])
            true_positives[cat_idx] += ((eval_img["dtMatches"] > 0) & dt_not_ignore).sum(1)
            false_positives[cat_idx] += ((eval_img["dtMatches"] == 0) & dt_not_ignore).sum(1)
            false_negatives[cat_idx] += (eval_img["gtMatches"][:, gt_not_ignore_idxs] == 0).sum(1)
    return true_positives.astype(int), false_positives.astype(int), false_negatives.astype(int)


def get_counts_and_scores(cocoEval: SlyCOCOeval, cat_id: int, t: int):
    """Returns tps, fps, scores, n_positives"""

    if not pycocotools_installed:
        raise ImportError("pycocotools is not installed")

    aRng = cocoEval.params.areaRng[0]
    eval_imgs = [ev for ev in cocoEval.evalImgs if ev is not None and ev["aRng"] == aRng]

    tps = []
    fps = []
    # fns = []
    scores = []
    n_positives = 0

    # Process each evaluated image
    for eval_img in eval_imgs:
        if eval_img["category_id"] != cat_id:
            continue
        dtScores = eval_img["dtScores"]
        dtm = eval_img["dtMatches"][t]
        gtm = eval_img["gtMatches"][t]

        # ntp = (dtm > 0).sum()
        # nfp = (dtm == 0).sum()
        # nfn = (gtm == 0).sum()
        p = len(gtm)

        tp = (dtm > 0).astype(int).tolist()
        fp = (dtm == 0).astype(int).tolist()
        # fn = [nfn]*len(dtm)

        tps.extend(tp)
        fps.extend(fp)
        # fns.extend(fn)
        scores.extend(dtScores)
        n_positives += p

    assert len(tps) == len(fps) == len(scores)

    # sort by score
    indices = np.argsort(scores)[::-1]
    scores = np.array(scores)[indices]
    tps = np.array(tps)[indices]
    fps = np.array(fps)[indices]

    return tps, fps, scores, n_positives


def get_eval_img_dict(cocoEval: SlyCOCOeval):
    if not pycocotools_installed:
        raise ImportError("pycocotools is not installed")

    aRng = cocoEval.params.areaRng[0]
    eval_img_dict = defaultdict(list)  # img_id : dt/gt
    for i, eval_img in enumerate(cocoEval.evalImgs):
        if eval_img is None or eval_img["aRng"] != aRng:
            continue
        img_id = eval_img["image_id"]
        cat_id = eval_img["category_id"]
        ious = cocoEval.ious[(img_id, cat_id)]
        # ! inplace operation
        eval_img["ious"] = ious
        eval_img_dict[img_id].append(eval_img)
    eval_img_dict = dict(eval_img_dict)
    return eval_img_dict


def _get_missclassified_match(eval_img_cls: SlyCOCOeval, dt_id, gtIds_orig, dtIds_orig, iou_t):
    if not pycocotools_installed:
        raise ImportError("pycocotools is not installed")

    # Correction on miss-classification
    gt_idx = np.nonzero(eval_img_cls["gtMatches"][iou_t] == dt_id)[0]
    if len(gt_idx) == 1:
        gt_idx = gt_idx[0]
        gt_id = eval_img_cls["gtIds"][gt_idx]
        gt_idx_o = gtIds_orig.index(gt_id)
        dt_idx_o = dtIds_orig.index(dt_id)
        iou = eval_img_cls["ious"][dt_idx_o, gt_idx_o].item()
        is_gt_ignore = eval_img_cls["gtIgnore"][gt_idx]
        if not is_gt_ignore:
            return gt_id, iou
    elif len(gt_idx) > 1:
        raise ValueError("Multiple matches")
    return None, None


def get_matches(
    eval_img_dict: dict,
    eval_img_dict_cls: dict,
    cocoEval_cls: SlyCOCOeval,
    iou_idx_per_class: dict = None,
):
    if not pycocotools_installed:
        raise ImportError("pycocotools is not installed")

    cat_ids = cocoEval_cls.cocoGt.getCatIds()
    matches = []
    for img_id, eval_imgs in eval_img_dict.items():

        # get miss-classified
        eval_img_cls = eval_img_dict_cls[img_id][0]
        gt_ids_orig_cls = [_["id"] for i in cat_ids for _ in cocoEval_cls._gts[img_id, i]]

        for eval_img in eval_imgs:
            cat_id = eval_img["category_id"]
            iou_t = iou_idx_per_class[cat_id]
            dtIds = np.array(eval_img["dtIds"])
            gtIds = np.array(eval_img["gtIds"])
            dtm = eval_img["dtMatches"][iou_t]
            gtm = eval_img["gtMatches"][iou_t]
            dtIgnore = eval_img["dtIgnore"][iou_t]
            gtIgnore = eval_img["gtIgnore"]

            # True Positives
            tp_idxs = np.nonzero(dtm)[0]
            for i in tp_idxs:
                if dtIgnore[i]:
                    continue
                dt_id = dtIds[i]
                gt_id = int(dtm[i])
                gt_idx = np.where(gtIds == gt_id)[0]
                iou = eval_img["ious"][i, gt_idx].item()
                score = eval_img["dtScores"][i]
                match = {
                    "image_id": eval_img["image_id"],
                    "category_id": eval_img["category_id"],
                    "dt_id": dt_id,
                    "gt_id": gt_id,
                    "type": "TP",
                    "score": score,
                    "iou": iou,
                    "miss_cls": False,
                }
                assert iou >= 0.5, iou
                matches.append(match)

            # False Positives
            fp_idxs = np.nonzero(dtm == 0)[0]
            for i in fp_idxs:
                dt_id = dtIds[i]
                score = eval_img["dtScores"][i]
                match = {
                    "image_id": eval_img["image_id"],
                    "category_id": eval_img["category_id"],
                    "dt_id": dt_id,
                    "gt_id": None,
                    "type": "FP",
                    "score": score,
                    "iou": None,
                    "miss_cls": False,
                }

                # Correction on miss-classification
                cls_gt_id, iou = _get_missclassified_match(
                    eval_img_cls, dt_id, gt_ids_orig_cls, eval_img_cls["dtIds"], iou_t
                )
                if cls_gt_id is not None:
                    assert iou >= 0.5, iou
                    match["gt_id"] = cls_gt_id
                    match["iou"] = iou
                    match["miss_cls"] = True
                matches.append(match)

            # False Negatives
            fn_idxs = np.nonzero(gtm == 0)[0]
            for i in fn_idxs:
                if gtIgnore[i]:
                    continue
                gt_id = gtIds[i]
                match = {
                    "image_id": eval_img["image_id"],
                    "category_id": eval_img["category_id"],
                    "dt_id": None,
                    "gt_id": gt_id,
                    "type": "FN",
                    "score": None,
                    "iou": None,
                    "miss_cls": False,
                }
                matches.append(match)

    return matches


def get_rare_classes(cocoGt: COCO, topk_ann_fraction=0.1, topk_classes_fraction=0.2):
    if not pycocotools_installed:
        raise ImportError("pycocotools is not installed")

    anns_cat_ids = [ann["category_id"] for ann in cocoGt.anns.values()]
    cat_ids, cat_counts = np.unique(anns_cat_ids, return_counts=True)
    inds_sorted = np.argsort(cat_counts)
    cum_counts = cat_counts[inds_sorted].cumsum()

    topk_upper1 = np.sum(cum_counts < cum_counts[-1] * topk_ann_fraction)
    topk_upper2 = int(topk_classes_fraction * len(cat_counts))
    topk_lower1 = 1

    topk = max(topk_lower1, min(topk_upper1, topk_upper2))
    cat_ids_rare = cat_ids[inds_sorted][:topk]
    cat_names = [cocoGt.cats[cat_id]["name"] for cat_id in cat_ids_rare]
    return cat_ids_rare, cat_names



from typing import Dict, List, Optional

from fastapi import BackgroundTasks, Request

import supervisely.io.fs as sly_fs
from supervisely._utils import logger
from supervisely.api.api import Api
from supervisely.app.fastapi.subapp import Application
from supervisely.nn.inference.predict_app.gui.gui import PredictAppGui
from supervisely.nn.inference.predict_app.gui.utils import disable_enable
from supervisely.nn.model.prediction import Prediction


class PredictApp:
    def __init__(self, api: Api):
        _static_dir = "static"
        sly_fs.mkdir(_static_dir, True)
        self.api = api
        self.gui = PredictAppGui(api, static_dir=_static_dir)
        self.app = Application(self.gui.layout, static_dir=_static_dir)
        self._add_endpoints()

        @self.gui.output_selector.start_button.click
        def start_prediction():
            if self.gui.output_selector.validate_step():
                widgets_to_disable = self.gui.output_selector.widgets_to_disable + [self.gui.settings_selector.preview.run_button]
                disable_enable(widgets_to_disable, True)
                self.gui.run()
                self.shutdown_serving_app()
                self.shutdown_predict_app()

    def shutdown_serving_app(self):
        if self.gui.output_selector.should_stop_serving_on_finish():
            logger.info("Stopping serving app...")
            self.gui.model_selector.model.stop()

    def shutdown_predict_app(self):
        if self.gui.output_selector.should_stop_self_on_finish():
            self.gui.output_selector.start_button.disable()
            logger.info("Stopping Predict App...")
            self.app.stop()
        else:
            disable_enable(self.gui.output_selector.widgets_to_disable, False)
            self.gui.output_selector.start_button.enable()

    def run(self, run_parameters: Optional[Dict] = None) -> List[Prediction]:
        return self.gui.run(run_parameters)

    def stop(self):
        self.gui.stop()

    def shutdown_model(self):
        self.gui.shutdown_model()

    def load_from_json(self, data):
        self.gui.load_from_json(data)
        if data.get("run", False):
            try:
                self.run()
            except Exception as e:
                raise
            finally:
                if data.get("stop_after_run", False):
                    self.shutdown_model()
                    self.app.stop()

    def get_inference_settings(self):
        return self.gui.settings_selector.get_inference_settings()

    def get_run_parameters(self):
        return self.gui.get_run_parameters()

    def _add_endpoints(self):
        server = self.app.get_server()

        @server.post("/load")
        def load(request: Request, background_tasks: BackgroundTasks):
            """
            Load the model state from a JSON object.
            This endpoint initializes the model with the provided state.
            All the fields are optional

            Example state:
                state = {
                    "model": {
                        "mode": "connect",
                        "session_id": "12345"
                        # "mode": "pretrained",
                        # "framework: "YOLO",
                        # "model_name": "YOLO11m-seg",
                        # "mode": "custom",
                        # "train_task_id": 123
                    },
                    "input": {
                        "project_id": 123,
                        # "dataset_ids": [...],
                        # "video_id": 123
                    },
                    "settings": {
                        "inference_settings": {
                            "confidence_threshold": 0.5
                        },
                    }
                    "output": {
                        "mode": "create",
                        "project_name": "Predictions",
                        # "mode": "append",
                        # "mode": "replace",
                        # "mode": "iou_merge",
                        # "iou_merge_threshold": 0.5
                    }
                }
            """
            state = request.state.state
            stop_after_run = state.get("stop_after_run", False)
            if stop_after_run:
                state["stop_after_run"] = False
            self.load_from_json(state)
            if stop_after_run:
                self.shutdown_model()
                background_tasks.add_task(self.app.stop)

        @server.post("/deploy")
        def deploy(request: Request):
            """
            Deploy the model for inference.
            This endpoint prepares the model for running predictions.
            """
            self.gui.model_selector.model._deploy()

        @server.get("/inference_settings")
        def get_inference_settings():
            """
            Get the inference settings for the model.
            This endpoint returns the current inference settings.
            """
            return self.get_inference_settings()

        @server.get("/run_parameters")
        def get_run_parameters():
            """
            Get the run parameters for the model.
            This endpoint returns the parameters needed to run the model.
            """
            return self.get_run_parameters()

        @server.post("/predict")
        def predict(request: Request):
            """
            Run the model prediction.
            This endpoint processes the request data and runs the model prediction.

            Example data:
                data = {
                    "inference_settings": {
                        "conf": 0.6,
                    },
                    "input": {
                        # "project_id": ...,
                        # "dataset_ids": [...],
                        "image_ids": [1148679, 1148675],
                    },
                    "output": {"mode": "iou_merge", "iou_merge_threshold": 0.5},
                }
            """
            state = request.state.state
            predictions = self.run(state)
            return [prediction.to_json() for prediction in predictions]

        @server.post("/run")
        def run(request: Request):
            """
            Run the model prediction.
            """
            predicitons = self.run()
            return [prediction.to_json() for prediction in predicitons]



# coding: utf-8

# fields names
ID = 'id'
KEY = 'key'
# INDEX = 'index'
FIGURES = 'figures'
OBJECT_ID = 'objectId'
OBJECT_KEY = 'objectKey'
TAGS = 'tags'

TAGS_MAP = 'tags'
OBJECTS_MAP = 'objects'
FIGURES_MAP = 'figures'
POINTCLOUD_MAP = 'pointclouds'

OBJECTS = 'objects'
POINTCLOUD_ID = 'pointCloudId'
NAME = 'name'
TAGS = 'tags'
DESCRIPTION = 'description'



# coding: utf-8

DOWNLOAD_BATCH_SIZE = None

# for Video Labeling Toolbox
PLAYBACK_RATE_POSSIBLE_VALUES = [
    0.1,
    0.3,
    0.5,
    0.6,
    0.7,
    0.8,
    0.9,
    1,
    1.1,
    1.2,
    1.3,
    1.5,
    2,
    4,
    8,
    16,
    32,
]



# coding: utf-8

# fields names
ALPHA_MASK = "alpha_mask"
BITMAP = "bitmap"
MASK_3D = "mask_3d"
COLS = "cols"
DATA = "data"
EXTERIOR = "exterior"
FACES = "faces"
INTERIOR = "interior"
MULTICHANNEL_BITMAP = "multichannelBitmap"
ORIGIN = "origin"
SPACE_ORIGIN = "space_origin"
SPACE = "space"
SPACE_DIRECTIONS = "space_directions"
POINTS = "points"
ROWS = "rows"
TYPE = "type"
GEOMETRY_SHAPE = "shape"
GEOMETRY_TYPE = "geometryType"
ANY_SHAPE = "any"

NODES = "nodes"
EDGES = "edges"
ENABLED = "enabled"
LABEL = "label"
COLOR = "color"
TEMPLATE = "template"
LOCATION = "location"
LOC = "loc"

X = "x"
Y = "y"
Z = "z"
POSITION = "position"
ROTATION = "rotation"
DIMENTIONS = "dimensions"


LABELER_LOGIN = "labelerLogin"
UPDATED_AT = "updatedAt"
CREATED_AT = "createdAt"
ID = "id"
CLASS_ID = "classId"

INDICES = "indices"

TRACK_ID = "trackId"



import json
from pathlib import Path
from typing import Dict

from jinja2 import Template

from supervisely.io.fs import ensure_base_path
from supervisely.nn.benchmark.visualization.widgets.widget import BaseWidget


class ChartWidget(BaseWidget):
    def __init__(
        self,
        name: str,
        figure,  # plotly figure
        switchable: bool = False,
        switch_key: str = "switch_key",
        radiogroup_id: str = None,
    ) -> None:
        super().__init__(name)
        self.switchable = switchable
        self.switch_key = switch_key
        self.radiogroup_id = radiogroup_id

        self.figure = figure
        self.click_data = None
        self.click_gallery_id = None
        self.chart_click_extra = None

    def save_data(self, basepath: str) -> None:
        # init data
        basepath = basepath.rstrip("/")
        ensure_base_path(basepath + self.data_source)

        with open(basepath + self.data_source, "w") as f:
            json.dump(self.get_init_data(), f)

        # click data
        if self.click_data is not None:
            ensure_base_path(basepath + self.click_data_source)
            with open(basepath + self.click_data_source, "w") as f:
                json.dump(self.click_data, f)

    def set_click_data(
        self, gallery_id: str, click_data: Dict, chart_click_extra: str = ""
    ) -> None:
        self.click_data = click_data
        self.click_gallery_id = gallery_id
        self.chart_click_extra = chart_click_extra

    def _get_template_data(self):
        return {
            "widget_id": self.id,
            "radio_group": self.radiogroup_id,
            "switchable": self.switchable,
            "switch_key": self.switch_key,
            "init_data_source": self.data_source,
            "click_handled": self.click_data is not None,
            "chart_click_data_source": self.click_data_source,
            "gallery_id": self.click_gallery_id,
            "chart_click_extra": self.chart_click_extra,
        }

    def to_html(self) -> str:
        template_str = Path(__file__).parent / "template.html"
        return Template(template_str.read_text()).render(self._get_template_data())

    def get_init_data(self):
        return {
            "selected": None,
            "galleryContent": "",
            "dialogVisible": False,
            "chartContent": json.loads(self.figure.to_json()),
        }

    def get_state(self) -> Dict:
        res = {}
        if self.switchable:
            res[self.radiogroup_id] = self.switch_key
        return res



from typing import Any, Dict, List

from supervisely import logger
from supervisely.api.api import Api
from supervisely.app.widgets import (
    Button,
    Card,
    Checkbox,
    Container,
    Field,
    Input,
    OneOf,
    Progress,
    ProjectThumbnail,
    RadioGroup,
    Text,
)
from supervisely.project.project_meta import ProjectType


class OutputSelector:
    title = "Result"
    description = "Select the output mode"
    lock_message = "Select previous step to unlock"

    def __init__(self, api: Api):
        # Init Step
        self.api = api
        self.display_widgets: List[Any] = []
        # -------------------------------- #

        # Init Base Widgets
        self.validator_text = None
        self.start_button = None
        self.stop_button = None
        self.container = None
        self.card = None
        # -------------------------------- #

        # Init Step Widgets
        self.stop_serving_on_finish = None
        self.stop_self_on_finish = None
        self.project_name_input = None
        self.project_name_field = None
        self.progress = None
        self.project_thumbnail = None
        # -------------------------------- #

        # TODO: Implement option later
        # Stop Apps on Finish
        # self.stop_serving_on_finish = Checkbox("Stop Serving App on prediction finish", False)
        # self.stop_self_on_finish = Checkbox("Stop Predict App on prediction finish", True)
        # Add widgets to display ------------ #
        # self.display_widgets.extend([self.stop_serving_on_finish, self.stop_self_on_finish])
        # ----------------------------------- #

        # Project Name
        self.project_name_input = Input(minlength=1, maxlength=255, placeholder="New Project Name")
        self.project_name_field = Field(
            content=self.project_name_input,
            title="New Project Name",
            description="Name of the new project to create for the results. The created project will have the same dataset structure as the input project.",
        )
        self.skip_annotated_checkbox = Checkbox("Skip annotated items", False)
        self._tab_names = ["Create New Project", "Update source project"]
        self._tab_contents = [self.project_name_field, self.skip_annotated_checkbox]
        self.tabs = RadioGroup(
            items=[
                RadioGroup.Item(tab_name, content=tab_content)
                for tab_name, tab_content in zip(self._tab_names, self._tab_contents)
            ],
        )
        self.oneof = OneOf(self.tabs)
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.tabs, self.oneof])
        # ----------------------------------- #

        # Base Widgets
        self.validator_text = Text("", status="text")
        self.validator_text.hide()
        self.start_button = Button("Run", icon="zmdi zmdi-play")
        self.stop_button = Button("Stop", icon="zmdi zmdi-stop")
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.start_button, self.validator_text])
        # ----------------------------------- #

        # Progress
        self.progress = Progress(hide_on_finish=False)
        self.progress.hide()
        self.secondary_progress = Progress(hide_on_finish=False)
        self.secondary_progress.hide()
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.progress, self.secondary_progress])
        # ----------------------------------- #

        # Result
        self.project_thumbnail = ProjectThumbnail()
        self.project_thumbnail.hide()
        # Add widgets to display ------------ #
        self.display_widgets.extend([self.project_thumbnail])
        # ----------------------------------- #

        # Card Layout
        self.container = Container(self.display_widgets)
        self.card = Card(
            title=self.title,
            description=self.description,
            content=self.container,
            lock_message=self.lock_message,
        )
        # ----------------------------------- #

    def lock(self):
        self.card.lock(self.lock_message)

    def unlock(self):
        self.card.unlock()

    @property
    def widgets_to_disable(self) -> list:
        return [self.project_name_input]

    def set_result_thumbnail(self, project_id: int):
        try:
            project_info = self.api.project.get_info_by_id(project_id)
            self.project_thumbnail.set(project_info)
            self.project_thumbnail.show()
        except Exception as e:
            logger.error(f"Failed to set result thumbnail: {str(e)}")
            self.project_thumbnail.hide()

    def get_settings(self) -> Dict[str, Any]:
        settings = {}
        if self.tabs.get_value() == self._tab_names[1]:
            settings["upload_to_source_project"] = True
        else:
            settings["project_name"] = self.project_name_input.get_value()
        settings["skip_annotated"] = self.skip_annotated_checkbox.is_checked()
        return settings

    def should_stop_serving_on_finish(self) -> bool:
        if self.stop_serving_on_finish is not None:
            return self.stop_serving_on_finish.is_checked()
        return False

    def should_stop_self_on_finish(self) -> bool:
        if self.stop_self_on_finish is not None:
            return self.stop_self_on_finish.is_checked()
        return True

    def load_from_json(self, data):
        project_name = data.get("project_name", None)
        if project_name:
            self.project_name_input.set_value(project_name)
        upload_to_source_project = data.get("upload_to_source_project", False)
        if upload_to_source_project:
            self.tabs.set_value(self._tab_names[1])
        else:
            self.tabs.set_value(self._tab_names[0])

    def validate_step(self) -> bool:
        self.validator_text.hide()
        if (
            self.tabs.get_value() == self._tab_names[0]
            and self.project_name_input.get_value() == ""
        ):
            self.validator_text.set(text="Project name is required", status="error")
            self.validator_text.show()
            return False

        return True

    def update_item_type(self, item_type: str):
        if item_type == ProjectType.IMAGES.value:
            self.skip_annotated_checkbox.show()
        elif item_type == ProjectType.VIDEOS.value:
            self.skip_annotated_checkbox.hide()
        else:
            raise ValueError(f"Unsupported item type: {item_type}")



from supervisely.nn.training.loggers.base_train_logger import BaseTrainLogger
from supervisely.nn.training.loggers.tensorboard_logger import (
    TensorboardLogger,
    tensorboard_installed,
)

train_logger = BaseTrainLogger()


def setup_train_logger(name="tensorboard_logger"):
    global train_logger
    if name == "tensorboard_logger":
        if tensorboard_installed:
            if not isinstance(train_logger, TensorboardLogger):
                train_logger = TensorboardLogger()
        else:
            raise ImportError("TensorboardX is not installed")
    elif name == "default_logger":
        if not isinstance(train_logger, BaseTrainLogger):
            train_logger = BaseTrainLogger()
    else:
        raise ValueError(f"Logger {name} is not supported")
